#include <plugin.hpp>
#include <CRT/crt.hpp>
#include <DlgBuilder.hpp>
#include "TrueBranchClass.hpp"
#include "TrueBranch.hpp"
#include "TrueBranchLng.hpp"
#include "TrueOpt.hpp"
#include "version.hpp"
#include <initguid.h>
#include "guid.hpp"

//#include "_syslog.hpp"

#if defined(__GNUC__)

#ifdef __cplusplus
extern "C"{
#endif
  BOOL WINAPI DllMainCRTStartup(HANDLE hDll,DWORD dwReason,LPVOID lpReserved);
#ifdef __cplusplus
};
#endif

BOOL WINAPI DllMainCRTStartup(HANDLE hDll,DWORD dwReason,LPVOID lpReserved)
{
  (void) lpReserved;
  (void) dwReason;
  (void) hDll;
  return TRUE;
}
#endif

struct PluginStartupInfo Info;
struct FarStandardFunctions FSF;

#ifdef FAR2
int WINAPI GetMinFarVersionW()
{
  return FARMANAGERVERSION;
}
#else
void WINAPI GetGlobalInfoW(struct GlobalInfo *Info)
{
  Info->StructSize=sizeof(GlobalInfo);
  Info->MinFarVersion=FARMANAGERVERSION;
  Info->Version=PLUGIN_VERSION;
  Info->Guid=MainGuid;
  Info->Title=PLUGIN_NAME;
  Info->Description=PLUGIN_DESC;
  Info->Author=PLUGIN_AUTHOR;
}
#endif

void WINAPI SetStartupInfoW(const struct PluginStartupInfo *psi)
{
  Info = *psi;
  FSF = *psi->FSF;
  Info.FSF = &FSF;
  GetOpt();
}

void WINAPI GetPluginInfoW(struct PluginInfo *Info)
{
  static const TCHAR *PluginMenuStrings[1];
  static const TCHAR *PluginCfgStrings[1];
  static const TCHAR *DiskMenuStrings[1];

  memset(Info, 0, sizeof(struct PluginInfo));
  Info->StructSize = sizeof(struct PluginInfo);
#ifndef FAR2
  Info->PluginMenu.Guids=&MenuGuid;
  Info->PluginConfig.Guids=&MenuGuid;
#endif
  if ( Opt.Add2PlugMenu )
  {
    PluginMenuStrings[0] = GetMsg(MTitle);
#ifdef FAR2
    Info->PluginMenuStrings = PluginMenuStrings;
    Info->PluginMenuStringsNumber = ARRAYSIZE(PluginMenuStrings);
#else
    Info->PluginMenu.Strings=PluginMenuStrings;
    Info->PluginMenu.Count=ARRAYSIZE(PluginMenuStrings);
    Info->PluginMenu.Guids=&PluginMenuGuid;
#endif
  }
  if ( Opt.Add2DisksMenu )
  {
    DiskMenuStrings[0] = GetMsg(MTitle);
#ifdef FAR2
    Info->DiskMenuStrings = DiskMenuStrings;
    Info->DiskMenuStringsNumber = 1;
#else
    Info->DiskMenu.Strings = DiskMenuStrings;
    Info->DiskMenu.Count = 1;
    Info->DiskMenu.Guids = &DiskMenuGuid;
#endif
  }
  PluginCfgStrings[0] = GetMsg(MTitle);
#ifdef FAR2
  Info->PluginConfigStrings = PluginCfgStrings;
  Info->PluginConfigStringsNumber = ARRAYSIZE(PluginCfgStrings);
#else
  Info->PluginConfig.Strings=PluginCfgStrings;
  Info->PluginConfig.Count=ARRAYSIZE(PluginCfgStrings);
  Info->PluginConfig.Guids=&PluginConfigGuid;
#endif
}

static const TCHAR HstFiles[] = _T("TrueBranch_FilesMask");

#ifdef FAR2
int WINAPI ConfigureW( int)
#else
intptr_t WINAPI ConfigureW( const ConfigureInfo*)
#endif
{
  PluginDialogBuilder Builder(
    Info,
#ifndef FAR2
    MainGuid, DialogGuid,
#endif
    MTitle, HlfId.Config
  );
  Builder.AddCheckbox(MAdd2PlugMenu,  &Opt.Add2PlugMenu);
  Builder.AddCheckbox(MAdd2DisksMenu, &Opt.Add2DisksMenu);
  Builder.AddSeparator();
  Builder.AddCheckbox(MESCConfirm,    &Opt.ESCConfirm);
  Builder.AddCheckbox(MWarnGetInfo,   &Opt.WarnGetInfo);
  Builder.AddCheckbox(MEAutoSelected, &Opt.AutoSelected);
  Builder.AddSeparator();
  Builder.AddCheckbox(MShowFolders,   &Opt.ShowFolders, 0, true);
  Builder.AddCheckbox(MScanSymlink,   &Opt.ScanSymlink);
  Builder.AddCheckbox(MFullPath,      &Opt.FullPath,    0, true);
  Builder.AddSeparator();
  Builder.AddText(MFilesMask);
  Builder.AddEditField(Opt.FilesMask, ARRAYSIZE(Opt.FilesMask), 40, HstFiles);
  Builder.AddOKCancel(MOk, MCancel);
  if ( Builder.ShowDialog() )
  {
    parseMasks();
    SetOpt(enoAll);
    return true;
  }
  return false;
}

static int scanRoot(struct PanelInfo *pi, bool Selected, TRootItem **rootList, const TCHAR* CurDir, int *n)
{
//SysLog(_T("  scanRoot\r\n  {\r\n"));
  int subfolders = 0;
  for ( size_t i = 0 ; i < (size_t)( Selected ? pi->SelectedItemsNumber : pi->ItemsNumber ) ; i++ )
  {
    size_t sz = Info.PanelControl(PANEL_ACTIVE, Selected ? FCTL_GETSELECTEDPANELITEM : FCTL_GETPANELITEM, i, 0);
    if ( sz )
    {
      PluginPanelItem* PPI = (PluginPanelItem*)malloc(sz);
      if ( PPI )
      {
#ifdef FAR2
        Info.PanelControl(PANEL_ACTIVE, Selected ? FCTL_GETSELECTEDPANELITEM : FCTL_GETPANELITEM, i, (PANELCONTROLPARAM2)PPI);
#else
        FarGetPluginPanelItem FPPI={sizeof(FarGetPluginPanelItem), sz, PPI};
        Info.PanelControl(PANEL_ACTIVE, Selected ? FCTL_GETSELECTEDPANELITEM : FCTL_GETPANELITEM, i, (PANELCONTROLPARAM2)&FPPI);
#endif
//SysLog(_T("    '%s'\r\n"), piGetFileName(PPI));
        if ( lstrcmp(piGetFileName(PPI), _T("..")) )
        {
          bool Dir = ( piGetFileAttributes(PPI) & FILE_ATTRIBUTE_DIRECTORY ) ? true : false;
          if ( !Selected || PPI->Flags & PPIF_SELECTED )
          {
            if ( Dir )
              subfolders++;
            if ( rootList )
              add2RootList(rootList, (*n)++, Dir, piGetFileName(PPI), CurDir);
          }
        }
        free(PPI);
      }
    }
  }
//SysLog(_T("  }\r\n"));
  return subfolders;
}

BOOL ConfigAdv(void)
{
  PluginDialogBuilder Builder(
    Info,
#ifndef FAR2
    MainGuid, DialogGuid,
#endif
    MTitleAdv, HlfId.ConfigAdv
  );
  Builder.AddCheckbox(MShowFolders,  &Opt.ShowFolders, 0, true);
  Builder.AddCheckbox(MScanSymlink,  &Opt.ScanSymlink);
  Builder.AddCheckbox(MFullPath,     &Opt.FullPath,    0, true);
  Builder.AddSeparator();
  Builder.AddText(MFilesMask);
  Builder.AddEditField(Opt.FilesMask, ARRAYSIZE(Opt.FilesMask), 40, HstFiles);
  Builder.AddOKCancel(MOk, MCancel);
  if ( Builder.ShowDialog() )
  {
    parseMasks();
    SetOpt(enoAll);
    return true;
  }
  return false;
}

#ifdef FAR2
#define OPENW_CANCEL INVALID_HANDLE_VALUE
HANDLE WINAPI OpenPluginW(int OpenFrom,INT_PTR item)
#else
#define OPENW_CANCEL NULL
HANDLE WINAPI OpenW(const struct OpenInfo *oInfo)
#endif
{
//SysLog(_T("OpenW\r\n{\r\n"));
  struct PanelInfo pi;
#ifndef FAR2
  pi.StructSize = sizeof(PanelInfo);
#endif
  if ( Info.PanelControl(PANEL_ACTIVE, FCTL_GETPANELINFO, 0, (PANELCONTROLPARAM2)&pi) )
  {
    DWORD difAdd = 0;
    const TCHAR *MsgItems[]={GetMsg(MTitle),GetMsg(MErrBadPanel1),GetMsg(MErrBadPanel2),GetMsg(MCancel)};
    switch ( pi.PanelType )
    {
      case PTYPE_TREEPANEL:
        difAdd = DIF_DISABLE;
        break;
      case PTYPE_FILEPANEL:
        if ( pi.Flags & PFLAGS_REALNAMES )
        {
          if ( !scanRoot(&pi, false, NULL, NULL, NULL) )
          {
            const TCHAR *MsgItems[]={GetMsg(MTitle),GetMsg(MErrSubfolders),GetMsg(MCancel)};
            INFOMESSAGE(FMSG_WARNING, MsgItems, 1);
            return OPENW_CANCEL;
          }
          difAdd = scanRoot(&pi, true, NULL, NULL, NULL) ? 0 : MIF_DISABLE;
        }
        else
        {
          INFOMESSAGE(FMSG_WARNING, MsgItems, 1);
          return OPENW_CANCEL;
        }
        break;
      default:
        INFOMESSAGE(FMSG_WARNING, MsgItems, 1);
        return OPENW_CANCEL;
    }
    if ( difAdd )
      Opt.ScanMode = 0;

    struct ex(FarMenuItem) shMenu[4];
    memset(&shMenu, 0, sizeof(shMenu));
    shMenu[0].Text = GetMsg(MScanCurrent);
    shMenu[1].Text = GetMsg(MScanSelected);
    shMenu[1].Flags = difAdd;
    shMenu[2].Flags = MIF_SEPARATOR;
    shMenu[3].Text = GetMsg(MConfigAdv);

    BOOL ret = FALSE;
    int ExitCode;
    if ( Opt.AutoSelected )
      ExitCode = difAdd ? 0 : 1;
    else
      ExitCode = Opt.ScanMode;
    for ( ; ; )
    {
      for ( int i = 0 ; i < ARRAYSIZE(shMenu) ; i++ )
        if ( i == ExitCode )
          shMenu[i].Flags |= MIF_SELECTED;
        else
          shMenu[i].Flags &= ~MIF_SELECTED;
      ExitCode = INFOMENU(-1,-1,0,FMENU_WRAPMODE,GetMsg(MScanTitle),NULL,HlfId.Init,NULL,NULL,shMenu,ARRAYSIZE(shMenu));
      if ( ExitCode == 0 || ExitCode == 1  )
      {
        Opt.ScanMode = ExitCode;
        SetOpt(enoScanMode);
        ret = TRUE;
        break;
      }
      if ( ExitCode == 3 )
      {
        ConfigAdv();
        continue;
      }
      break;
    }
    if ( ret )
    {
      TCHAR CurDir[512] = _T("");
      if ( !piPlugin(pi) )
        FSF.GetCurrentDirectory(512, CurDir);
//SysLog(_T("  CurDir = '%s'\r\n"), CurDir);
      TRootItem *rootList = NULL;
      int rootSize = 0;
//SysLog(_T("  ScanMode = %d\r\n"), Opt.ScanMode);
      if ( Opt.ScanMode == 0 )
        if ( piPlugin(pi) )
          scanRoot(&pi, false, &rootList, CurDir, &rootSize);
        else
          add2RootList(&rootList, rootSize++, true, _T(""), CurDir);
      else
        scanRoot(&pi, true, &rootList, CurDir, &rootSize);
      if ( rootSize )
      {
//SysLog(_T("  rootSize = %d\r\n"), rootSize);
        HANDLE hPlugin = new BranchPanel(rootSize, rootList);
        if ( hPlugin == NULL )
          return OPENW_CANCEL;
//SysLog(_T("} = 1\r\n"));
        return hPlugin;
      }
    }
  }
//SysLog(_T("} = 0\r\n"));
  return OPENW_CANCEL;
}

#ifdef FAR2
int WINAPI GetFindDataW(HANDLE hPlugin,struct PluginPanelItem **pPanelItem,int *pItemsNumber,int OpMode)
{
  BranchPanel *Panel=(BranchPanel *)hPlugin;
  return Panel->GetFindData(pPanelItem,(size_t*)pItemsNumber,OpMode);
}
#else
intptr_t WINAPI GetFindDataW(struct GetFindDataInfo *gInfo)
{
  BranchPanel *Panel=(BranchPanel *)gInfo->hPanel;
  return Panel->GetFindData(&gInfo->PanelItem,&gInfo->ItemsNumber,gInfo->OpMode);
}
#endif

#ifdef FAR2
void WINAPI GetOpenPluginInfoW(HANDLE hPlugin, struct OpenPluginInfo *oInfo)
{
  BranchPanel *Panel=(BranchPanel *)hPlugin;
  Panel->GetOpenPanelInfo(oInfo);
}
#else
void WINAPI GetOpenPanelInfoW(struct OpenPanelInfo *oInfo)
{
  BranchPanel *Panel=(BranchPanel *)oInfo->hPanel;
  Panel->GetOpenPanelInfo(oInfo);
}
#endif

#ifdef FAR2
int WINAPI SetDirectoryW(HANDLE hPlugin,const TCHAR *Dir,int OpMode)
{
  BranchPanel *Panel=(BranchPanel *)hPlugin;
  return(Panel->SetDirectory(Dir,OpMode));
}
#else
intptr_t WINAPI SetDirectoryW(const struct SetDirectoryInfo *sInfo)
{
  BranchPanel *Panel=(BranchPanel *)sInfo->hPanel;
  return(Panel->SetDirectory(sInfo->Dir,sInfo->OpMode));
}
#endif

#ifdef FAR2
int WINAPI PutFilesW(HANDLE hPlugin,struct PluginPanelItem *PanelItem, int ItemsNumber,int Move,const wchar_t* SrcPath,int OpMode)
{
  BranchPanel *Panel=(BranchPanel *)hPlugin;
  return Panel->PutFiles(PanelItem,ItemsNumber,Move,SrcPath,OpMode);
}
#else
intptr_t WINAPI PutFilesW(const struct PutFilesInfo *pInfo)
{
  BranchPanel *Panel=(BranchPanel *)pInfo->hPanel;
  return Panel->PutFiles(pInfo->PanelItem,pInfo->ItemsNumber,pInfo->Move,pInfo->SrcPath,pInfo->OpMode);
}
#endif

#ifdef FAR2
void WINAPI ClosePluginW(HANDLE hPlugin)
{
  delete (BranchPanel *)hPlugin;
}
#else
void WINAPI ClosePanelW(const struct ClosePanelInfo *cInfo)
{
  delete (BranchPanel *)cInfo->hPanel;
}
#endif

#ifdef FAR2
int WINAPI ProcessKeyW(HANDLE hPlugin,int Key,unsigned int ControlState)
{
  BranchPanel *Panel=(BranchPanel *)hPlugin;
  return Panel->ProcessKey(Key,ControlState);
}
#else
intptr_t WINAPI ProcessPanelInputW(const struct ProcessPanelInputInfo *Info)
{
  BranchPanel *Panel=(BranchPanel *)(Info->hPanel);
  return Panel->ProcessKey(&Info->Rec);
}
#endif
