#include "plugin.hpp"
#include "CRT/crt.hpp"
#include <DlgBuilder.hpp>
#include "TrueBranch.hpp"
#include "TrueBranchClass.hpp"
#include "TrueOpt.hpp"
#include "TrueBranchLng.hpp"

//#include "_syslog.hpp"

BranchPanel::BranchPanel(int aRootItemsNumber, TRootItem *aRootList)
{
  rootList = aRootList;
  rootItemsNumber = aRootItemsNumber;
  BranchPanelItem = NULL;
  currRoot = NULL;
  alreadyInUpdate = false;
  pressedEsc = BranchItemsNum = BranchItemsMax = 0;
  calcRoot();
}

static void FreePanelItems(PluginPanelItem *Items, DWORD Total)
{
  if ( Items )
  {
    for ( DWORD I = 0 ; I < Total ; I++ )
    {
      if ( Items[I].Owner )
      {
        free((void*)Items[I].Owner);
        Items[I].Owner = NULL;
      }
      if ( piGetFileName(Items+I) )
      {
        free((void*)piGetFileName(Items+I));
        piGetFileName(Items+I) = NULL;
      }
    }
    free(Items);
  }
}

BranchPanel::~BranchPanel()
{
  free(rootList);
  FreePanelItems(BranchPanelItem, BranchItemsMax);
}

int BranchPanel::GetFindData(PluginPanelItem **pPanelItem,size_t *pItemsNumber,OPERATION_MODES OpMode)
{
//SysLog(_T("GetFindData\r\n"));
  if ( UpdateItems() )
  {
    *pPanelItem = BranchPanelItem;
    *pItemsNumber = BranchItemsNum;
  }
  else
  {
    *pPanelItem = NULL;
    *pItemsNumber = 0;
  }
  return TRUE;
}

void BranchPanel::GetOpenPanelInfo(struct
#ifdef FAR2
  OpenPluginInfo
#else
  OpenPanelInfo
#endif
*oInfo)
{
  oInfo->StructSize = sizeof(*oInfo);
  oInfo->Flags = OPIF_ADDDOTS|OPIF_REALNAMES
#ifdef FAR2
    |OPIF_USEFILTER|OPIF_USESORTGROUPS|OPIF_USEHIGHLIGHTING
#endif
  ;
  if ( !Opt.FullPath )
    oInfo->Flags |= OPIF_SHOWNAMESONLY;
  oInfo->HostFile = NULL;
  oInfo->CurDir = Opt.FullPath == 2 ? rootDir : _T("");
  oInfo->Format = GetMsg(MTitle);

  static TCHAR Title[256], tmp1[256], tmp2[512];
  lstrcpy(tmp1, GetMsg(MPanelTitle));
  if ( pressedEsc )
    lstrcat(tmp1, GetMsg(MUncomplete));
  int activeNumber = *tmp2 = 0;
  for ( int i = 0 ; i < rootItemsNumber ; i++ )
    if ( rootList[i].State == riActive )
  {
    if ( !activeNumber )
      lstrcpy(tmp2, rootList[i].Name);
    activeNumber++;
  }
  if ( activeNumber > 1 )
    FSF.sprintf(tmp2, GetMsg(MPanelTitleSelected), activeNumber);
  else
  {
    PanelInfo thisPInfo;
#ifndef FAR2
    thisPInfo.StructSize = sizeof(PanelInfo);
#endif
    Info.PanelControl(this, FCTL_GETPANELINFO, 0,(PANELCONTROLPARAM2)&thisPInfo);
    int len = (thisPInfo.PanelRect.right-thisPInfo.PanelRect.left)-lstrlen(tmp1)-8;
    if ( len < 10 )
      len = 10;
    if ( lstrlen(tmp2) > len )
      FSF.TruncPathStr(tmp2, len);
  }
  lstrcat(lstrcat(lstrcpy(Title, tmp1), tmp2), _T(" "));
  oInfo->PanelTitle = Title;
#ifdef FAR2
  static struct KeyBarTitles KeyBar;
  memset(&KeyBar,0,sizeof(KeyBar));
  KeyBar.Titles[7-1]=(TCHAR*)GetMsg(MF7);
  KeyBar.AltShiftTitles[3-1]=(TCHAR*)GetMsg(MAltShiftF3);
  KeyBar.AltShiftTitles[5-1]=(TCHAR*)GetMsg(MAltShiftF5);
  KeyBar.AltShiftTitles[6-1]=(TCHAR*)GetMsg(MAltShiftF6);
  KeyBar.AltShiftTitles[7-1]=(TCHAR*)GetMsg(MAltShiftF7);
#else
  static WORD FKeys[]=
  {
    VK_F7,0,MF7,
    VK_F3,SHIFT_PRESSED|LEFT_ALT_PRESSED,MAltShiftF3,
    VK_F5,SHIFT_PRESSED|LEFT_ALT_PRESSED,MAltShiftF5,
    VK_F6,SHIFT_PRESSED|LEFT_ALT_PRESSED,MAltShiftF6,
    VK_F7,SHIFT_PRESSED|LEFT_ALT_PRESSED,MAltShiftF7
  };
  static struct KeyBarLabel kbl[ARRAYSIZE(FKeys)/3];
  static struct KeyBarTitles KeyBar = {ARRAYSIZE(kbl), kbl};
  size_t i, j;
  for ( j = 0, i = 0 ; i < ARRAYSIZE(FKeys) ; i += 3 , ++j )
  {
    kbl[j].Key.VirtualKeyCode = FKeys[i];
    kbl[j].Key.ControlKeyState = FKeys[i+1];
    kbl[j].Text = kbl[j].LongText = FKeys[i+2] ? GetMsg(FKeys[i+2]) : L"";
  }
#endif
  oInfo->KeyBar=&KeyBar;
}

#ifdef FAR2
#define FCTL_CLOSEPANEL FCTL_CLOSEPLUGIN
#endif

int BranchPanel::SetDirectory(const TCHAR *Dir, OPERATION_MODES OpMode)
{
//SysLog(_T("SetDirectory(%s)\r\n"), Dir);
  if ( OpMode & OPM_FIND )
    return FALSE;
//  if ( !lstrcmp(Dir, _T("..")) )
//    return FALSE;
  Info.PanelControl(this, FCTL_CLOSEPANEL, 0, (PANELCONTROLPARAM2)(!lstrcmp(Dir, _T("..")) || !lstrcmp(Dir, _T("\\")) ? NULL : Dir));
  return TRUE;
}

static TCHAR foundDir[512] = _T("");

BOOL CheckForKey(WORD vKey = VK_ESCAPE)
{
  INPUT_RECORD *InputRec;
  DWORD n, ReadCnt;
  BOOL result = FALSE;
  HANDLE Console = GetStdHandle(STD_INPUT_HANDLE);

  if ( GetNumberOfConsoleInputEvents(Console, &n) )
  {
    if ( ( InputRec = (INPUT_RECORD *)malloc(n*sizeof(INPUT_RECORD)) ) != NULL )
    {
      if ( PeekConsoleInput(Console,InputRec,n,&ReadCnt) )
        if ( ReadCnt <= n )
          for ( DWORD i = 0 ; i < n ; i++ )
          {
            if ( InputRec[i].EventType == KEY_EVENT &&
                 InputRec[i].Event.KeyEvent.bKeyDown &&
                 InputRec[i].Event.KeyEvent.wVirtualKeyCode == vKey )
            {
              result=TRUE;
              if ( i > 0 )
                ReadConsoleInput(Console,InputRec,i,&ReadCnt);
              break;
            }
          }
      free(InputRec);
    }
  }
  return result;
}

static bool CheckForEsc(const TCHAR *FullName)
{
  static DWORD dwTicks = 0;
  DWORD dwNewTicks = GetTickCount();
  if (dwNewTicks - dwTicks < 1000)
    return false;
  TCHAR newDir[512];
  TCHAR *p = (TCHAR*)FSF.PointToName(lstrcpy(newDir, FullName));
  if ( p )
  {
    *p = 0;
    if ( FSF.LStricmp(foundDir, newDir) )
    {
      TCHAR truncDir[512];
      FSF.TruncPathStr(lstrcpy(truncDir, lstrcpy(foundDir, newDir)), lstrlen(GetMsg(MFindText)));
      const TCHAR *MsgItems[] = { GetMsg(MTitle), GetMsg(MFindTitle), truncDir, GetMsg(MFindText) };
      INFOMESSAGE(0, MsgItems, 0);
    }
  }
  dwTicks = dwNewTicks;
  if ( CheckForKey() )
  {
    if ( Opt.ESCConfirm )
    {
      const TCHAR *MsgItems[] = { GetMsg(MTitle), GetMsg(MEsc), GetMsg(MAbort), GetMsg(MContinue) };
      if ( INFOMESSAGE(FMSG_WARNING, MsgItems, 2) != 0 )
        return false;
    }
    return true;
  }
  return false;
}

int WINAPI findFunc(
#ifdef FAR2
  const FAR_FIND_DATA *ff,
#else
  const struct PluginPanelItem *ff,
#endif
  const TCHAR *FullName, void *Param
)
{
  if ( CheckForEsc(FullName) )
  {
    ((BranchPanel*)Param)->pressedEsc = 1;
    return FALSE;
  }
  if ( (
#ifdef FAR2
    ff->dwFileAttributes
#else
    ff->FileAttributes
#endif
    & FILE_ATTRIBUTE_DIRECTORY ) && !Opt.ShowFolders
  )
    return TRUE;
  if ( ((BranchPanel*)Param)->AddItem(ff, FullName) )
    return TRUE;
  return FALSE;
}

bool BranchPanel::AddDir(const TCHAR *dir)
{
//SysLog(_T("AddDir(%s)\r\n"), dir);
  DWORD Flags = FRS_RECUR;
  if ( Opt.ScanSymlink )
    Flags |= FRS_SCANSYMLINK;
  if ( Opt.FarSearch )
    FSF.FarRecursiveSearch(dir, Opt.FilesMask, findFunc, Flags, this);
  else
    FSF.FarRecursiveSearch(dir, _T("*"), findFunc, Flags, this);
  return true;
}

static bool warnGetInfo(const TCHAR *file)
{
  if ( Opt.WarnGetInfo )
  {
    TCHAR f[512];
    FSF.TruncPathStr(lstrcpy(f, file), 50);
    PluginDialogBuilder Builder(
      Info,
#ifndef FAR2
      MainGuid, DialogGuid,
#endif
      MTitle, HlfId.WarnGetInfo
    );
    Builder.AddText(MInfo);
    Builder.AddText(f);
    Builder.AddSeparator();
    Builder.AddCheckbox(MDontShow, &Opt.WarnGetInfo);
    Builder.AddOKCancel(MOk, MCancel);
    if ( Builder.ShowDialog() )
      SetOpt(enoWarnGetInfo);
  }
  return false;
}

static bool fIncl = false;
static bool fExcl = false;

static inline int matched(const TCHAR *mask, TCHAR *file)
{
  return FSF.ProcessName(mask, file, 0, PN_CMPNAMELIST);
}

bool BranchPanel::AddItem(
#ifdef FAR2
  const FAR_FIND_DATA *ff,
#else
  const struct PluginPanelItem *ff,
#endif
  const TCHAR *file, bool check
)
{
//SysLog(_T("AddItem(%s)\r\n"), file);
  bool add = true;
  if ( check )
  {
    DWORD dwAttr;
    if ( ff )
      dwAttr = ffGetFileAttributes(ff);
    else
    {
      dwAttr = GetFileAttributes(file);
      if ( dwAttr == INVALID_FILE_ATTRIBUTES )
        return warnGetInfo(file);
    }
    if ( dwAttr & FILE_ATTRIBUTE_DIRECTORY )
    {
//SysLog(_T("  Dir %s\r\n"), file);
      if ( Opt.ShowFolders != 1 )
        add = false;
    }
    else if ( !Opt.FarSearch )
    {
      TCHAR *fn = (TCHAR*)file;
      if ( currRoot )
      {
        int len = lstrlen(currRoot);
        if ( !FSF.LStrnicmp(currRoot, file, len) )
          fn += len;
      }
      if ( fIncl && !matched(Opt.IncludeMask, fn) )
        add = false;
      else if ( fExcl && matched(Opt.ExcludeMask, fn))
        add = false;
    }
  }
  if ( add )
  {
    if ( BranchItemsNum >= BranchItemsMax )
    {
      if ( BranchItemsMax > 1 )
        BranchItemsMax *= 2;
      else
        BranchItemsMax = 256;
      struct PluginPanelItem *NewPanelItem = (struct PluginPanelItem *)realloc(BranchPanelItem,sizeof(PluginPanelItem)*BranchItemsMax);
      if ( NewPanelItem == NULL )
      {
        const TCHAR *MsgItems[] = { GetMsg(MTitle), GetMsg(MMemory), GetMsg(MAbort) };
        INFOMESSAGE(FMSG_WARNING, MsgItems, 1);
        return false;
      }
      BranchPanelItem = NewPanelItem;
    }
    struct PluginPanelItem *CurPanelItem = &BranchPanelItem[BranchItemsNum];
    memset(CurPanelItem, 0, sizeof(*CurPanelItem));
    if ( !GetFileInfoAndValidate(file, ff, CurPanelItem) )
      return warnGetInfo(file);
    BranchItemsNum++;
  }
  return true;
}

static void WFD2FFD(WIN32_FIND_DATA &wfd, PluginPanelItem *ffd)
{
  piGetFileAttributes(ffd)    = wfd.dwFileAttributes;
  piGetCreationTime(ffd)      = wfd.ftCreationTime;
  piGetLastAccessTime(ffd)    = wfd.ftLastAccessTime;
  piGetLastWriteTime(ffd)     = wfd.ftLastWriteTime;
  piGetFileSize(ffd)          = ( (__int64)wfd.nFileSizeHigh << 32) | wfd.nFileSizeLow;
  piGetPackSize(ffd)          = piGetFileSize(ffd);
  piGetFileName(ffd)          = wcsdup(wfd.cFileName);
  piGetAlternateFileName(ffd) = NULL;
}

bool BranchPanel::GetFileInfoAndValidate(const TCHAR *FileName,
#ifdef FAR2
  const FAR_FIND_DATA *SrcData,
#else
  const struct PluginPanelItem *SrcData,
#endif
  struct PluginPanelItem *FindData
)
{
//SysLog(_T("GetFileInfoAndValidate(%s)\r\n"), FileName);
  bool ret = FALSE;
  if ( SrcData )
  {
    piGetFileAttributes(FindData)    = ffGetFileAttributes(SrcData);
    piGetCreationTime(FindData)      = ffGetCreationTime(SrcData);
    piGetLastAccessTime(FindData)    = ffGetLastAccessTime(SrcData);
    piGetLastWriteTime(FindData)     = ffGetLastWriteTime(SrcData);
    piGetFileSize(FindData)          = ffGetFileSize(SrcData);
    piGetPackSize(FindData)          = ffGetPackSize(SrcData);
    piGetFileName(FindData)          = wcsdup(FileName);
    piGetAlternateFileName(FindData) = NULL;
    ret = TRUE;
  }
  else if ( lstrlen(FileName) )
  {
    DWORD dwAttr = GetFileAttributes(FileName);
    if ( dwAttr != INVALID_FILE_ATTRIBUTES )
    {
      WIN32_FIND_DATA wfd;
      HANDLE fff = FindFirstFile(FileName, &wfd);
      if ( fff != INVALID_HANDLE_VALUE )
      {
        WFD2FFD(wfd, FindData);
        FindClose(fff);
      }
      else
      {
        wfd.dwFileAttributes = dwAttr;
        HANDLE hFile = CreateFile(FileName,FILE_READ_ATTRIBUTES,FILE_SHARE_READ|FILE_SHARE_WRITE|FILE_SHARE_DELETE,NULL,OPEN_EXISTING,FILE_FLAG_BACKUP_SEMANTICS|FILE_FLAG_POSIX_SEMANTICS,NULL);
        if ( hFile!=INVALID_HANDLE_VALUE )
        {
          GetFileTime(hFile, &wfd.ftCreationTime, &wfd.ftLastAccessTime, &wfd.ftLastWriteTime);
          wfd.nFileSizeLow = GetFileSize(hFile, &wfd.nFileSizeHigh);
          CloseHandle(hFile);
        }
        wfd.dwReserved0 = 0;
        wfd.dwReserved1 = 0;
        WFD2FFD(wfd, FindData);
      }
      if ( piGetFileName(FindData) )
        free((void*)piGetFileName(FindData));
      piGetFileName(FindData) = wcsdup(FileName);
      ret = TRUE ;
    }
  }
  if ( ret )
  {
    if ( piGetFileName(FindData) == NULL )
    {
      const TCHAR *MsgItems[] = { GetMsg(MTitle), GetMsg(MMemory), GetMsg(MAbort) };
      INFOMESSAGE(FMSG_WARNING, MsgItems, 1);
      ret = FALSE ;
    }
  }
  return ret;
}

int BranchPanel::UpdateItems(void)
{
//SysLog(_T("UpdateItems\r\n"));
  if ( alreadyInUpdate )
    return FALSE;
  alreadyInUpdate = true;
  HANDLE hScreen = Info.SaveScreen(0,0,-1,-1);
  *foundDir = 0;
  if ( BranchPanelItem )
  {
    FreePanelItems(BranchPanelItem, BranchItemsMax);
    BranchPanelItem = NULL;
  }
  BranchItemsNum = BranchItemsMax = 0;
  pressedEsc = 0;
  if ( !Opt.FarSearch )
  {
    fExcl = lstrcmp(Opt.ExcludeMask, _T("")) != 0;
    fIncl = lstrcmp(Opt.IncludeMask, _T("")) && !allFiles(Opt.IncludeMask);
  }
  for ( int i = 0 ; i < rootItemsNumber ; i++ )
    if ( rootList[i].State == riActive )
    {
      currRoot = rootList[i].Name;
      if ( rootList[i].Dir )
        AddDir(currRoot);
      else
        AddItem(NULL, currRoot);
      currRoot = NULL;
      if ( pressedEsc )
        break;
    }
  if ( BranchItemsNum && Opt.ShowFolders == 2 )
  {
    const TCHAR *MsgItems[] = { GetMsg(MTitle), GetMsg(MAddFolders) };
    INFOMESSAGE(0, MsgItems, 0);
    int BranchItemsBak = BranchItemsNum;
    for ( int i = 0 ; i < BranchItemsBak ; i++ )
    {
      TCHAR DirName[512];
      TCHAR *p = (TCHAR*)FSF.PointToName(lstrcpy(DirName, piGetFileName(BranchPanelItem+i)));
      if ( p > DirName )
      {
        *--p = 0;
        bool add = true;
        for ( int j = BranchItemsBak ; j < BranchItemsNum ; j++ )
          if ( !FSF.LStricmp(DirName, piGetFileName(BranchPanelItem+j)) )
          {
            add = false;
            break;
          }
        if ( add )
        {
          for ( int j = 0 ; j < rootItemsNumber ; j++ )
            if ( rootList[j].Dir && !FSF.LStricmp(DirName, rootList[j].Name) )
            {
              add = false;
              break;
            }
        }
        if ( add )
          AddItem(NULL, DirName, false);
      }
    }
  }
  if ( BranchItemsNum && *rootDir && Opt.FullPath == 2 )
  {
    int len = lstrlen(rootDir)+1;
    for ( int i = 0 ; i < BranchItemsNum ; i++ )
      lstrcpy((TCHAR*)piGetFileName(BranchPanelItem+i), piGetFileName(BranchPanelItem+i)+len);
  }
  const TCHAR *MsgItems[] = { GetMsg(MTitle), GetMsg(MUpdating) };
  INFOMESSAGE(0, MsgItems, 0);
  struct PluginPanelItem *NewPanelItem = (struct PluginPanelItem *)realloc(BranchPanelItem,sizeof(PluginPanelItem)*(BranchItemsMax = BranchItemsNum));
  if ( BranchItemsNum && NewPanelItem == NULL )
  {
    const TCHAR *MsgItems[] = { GetMsg(MTitle), GetMsg(MMemory), GetMsg(MAbort) };
    INFOMESSAGE(FMSG_WARNING, MsgItems, 1);
  }
  BranchPanelItem = NewPanelItem;
  Info.RestoreScreen(hScreen);
  alreadyInUpdate = false;
  return TRUE;
}

int BranchPanel::PutFiles(struct PluginPanelItem *PanelItem, int ItemsNumber, int, const TCHAR *SrcPath, OPERATION_MODES)
{
  int ret = 1;
  HANDLE hScreen = BeginPutFiles();
  for ( int i = 0 ; i < ItemsNumber ; i++ )
  {
    if ( !PutOneFile(SrcPath, PanelItem[i]) )
    {
      ret = FALSE;
      break;
    }
  }
  CommitPutFiles(hScreen);
  return ret;
}

HANDLE BranchPanel::BeginPutFiles()
{
  HANDLE hScreen = Info.SaveScreen(0,0,-1,-1);
  const TCHAR *MsgItems[] = { GetMsg(MTitle), GetMsg(MSendFiles) };
  INFOMESSAGE(0, MsgItems, 0);
  return hScreen;
}

void BranchPanel::CommitPutFiles(HANDLE hRestoreScreen)
{
  Info.RestoreScreen(hRestoreScreen);
}

void add2RootList(TRootItem **rootList, int n, bool Dir, const TCHAR *Name, const TCHAR* CurrentDir)
{
//SysLog(_T("    add2List(%s)\r\n"), Name);
  TRootItem *newList = (TRootItem *)realloc(*rootList, sizeof(TRootItem)*(n+1));
  newList[n].Dir = Dir;
  newList[n].State = riActive;
  lstrcpy(newList[n].Name, CurrentDir);
  if ( newList[n].Name[0] )
    FSF.AddEndSlash(newList[n].Name);
  if ( *Name )
    lstrcat(newList[n].Name, Name);
  else
  {
    TCHAR *p = newList[n].Name+(lstrlen(newList[n].Name)-1);
    if ( *p == _T('\\') )
      *p = 0;
  }
  *rootList = newList;
}

void BranchPanel::calcRoot(void)
{
  *rootDir = 0;
  if ( rootList )
  {
    int i0;
    for ( i0 = 0 ; i0 < rootItemsNumber ; i0++ )
      if ( rootList[i0].State == riActive )
      {
        lstrcpy(rootDir, rootList[i0].Name);
        break;
      }
    for ( int i = i0+1 ; i < rootItemsNumber ; i++ )
      if ( rootList[i].State == riActive )
      {
        TCHAR *p0 = rootDir, *p1 = rootList[i].Name;
        for ( ; *p0 && *p1 && *p0 == *p1 ; p0++, p1++ )
          ;
        if ( *p0 || *p1 )
        {
          *p0 = 0;
          while ( p0 > rootDir )
            if ( *--p0 == '\\' )
            {
              *p0 = 0;
              break;
            }
            else
              *p0 = 0;
        }
      }
  }
}

int BranchPanel::packRoot(void)
{
  int j = 0;
  int delCount = 0;
  for ( int i = 0 ; i < rootItemsNumber ; i++ )
  {
    if ( rootList[i].State == riDelete )
      delCount++;
    else
    {
      if ( i != j )
      {
        rootList[j].State = rootList[i].State;
        rootList[j].Dir = rootList[i].Dir;
        lstrcpy(rootList[j].Name, rootList[i].Name);
      }
      j++;
    }
  }
  return delCount;
}

int BranchPanel::PutOneFile(const TCHAR *SrcPath, PluginPanelItem &PanelItem)
{
  TRootItem newItem;
  TCHAR tmp[512], tmp2[512];
  newItem.State = riActive;
  newItem.Dir = ( piGetFileAttributes(&PanelItem) & FILE_ATTRIBUTE_DIRECTORY ) ? true : false;
  lstrcpy(newItem.Name, SrcPath);
  if ( *SrcPath )
    FSF.AddEndSlash(newItem.Name);
  lstrcat(newItem.Name, piGetFileName(&PanelItem));
  bool dupe = false;
  int delCount = 0;
  if ( newItem.Dir )
  {
    for ( int i = 0 ; i < rootItemsNumber ; i++ )
    {
      if ( rootList[i].State == riActive )
    {
        lstrcpy(tmp, newItem.Name);
        FSF.AddEndSlash(tmp);
        if ( rootList[i].Dir )
        {
          lstrcpy(tmp2, rootList[i].Name);
          FSF.AddEndSlash(tmp2);
          if ( !FSF.LStrnicmp(tmp, tmp2, lstrlen(tmp)) )
          {
            lstrcpy(rootList[i].Name, newItem.Name);
            dupe = true;
          }
          else if ( !FSF.LStrnicmp(tmp, tmp2, lstrlen(tmp2)) )
          {
            dupe = true;
            break;
          }
        }
        else
        {
          if ( !FSF.LStrnicmp(tmp, rootList[i].Name, lstrlen(tmp)) )
          {
            delCount++;
            rootList[i].State = riDelete;
          }
        }
    }
    }
    for ( int i = 0 ; i < rootItemsNumber ; i++ )
      if ( rootList[i].State != riDelete )
        for ( int j = i+1 ; j < rootItemsNumber ; j++ )
          if ( rootList[j].State != riDelete && !FSF.LStricmp(rootList[i].Name, rootList[j].Name ) )
          {
            delCount++;
            rootList[j].State = riDelete;
          }
  }
  else
  {
    for ( int i = 0 ; i < rootItemsNumber ; i++ )
    {
      if ( rootList[i].Dir )
      {
        lstrcpy(tmp, rootList[i].Name);
        FSF.AddEndSlash(tmp);
        if ( !FSF.LStrnicmp(tmp, newItem.Name, lstrlen(tmp)) )
          dupe = true;
      }
      else
      {
        if ( !FSF.LStricmp(rootList[i].Name, newItem.Name) )
          dupe = true;
      }
      if ( dupe )
        break;
    }
  }
  if ( delCount )
    rootItemsNumber -= packRoot();
  if ( !dupe )
  {
    add2RootList(&rootList, rootItemsNumber++, newItem.Dir, newItem.Name, _T(""));
    calcRoot();
  }
  return TRUE;
}

#ifdef FAR2
int BranchPanel::ProcessKey(int Key, unsigned int ControlState)
{
  if ( ! ( Key & PKF_PREPROCESS ) )
  {
    switch ( ControlState )
    {
      case 0:
        switch ( Key )
        {
          case VK_F1: Info.ShowHelp(Info.ModuleName, HlfId.Panel, FHELP_USECONTENTS|FHELP_NOSHOWERROR); return TRUE;
          case VK_F7: ProcessRemoveKey();                                                               return TRUE;
        }
        break;
      case PKF_SHIFT|PKF_ALT:
        switch ( Key )
        {
          case VK_F3: ProcessGoToKey();                     return TRUE;
          case VK_F5: ProcessToggleKey(Opt.FullPath, 3);    return TRUE;
          case VK_F6: ProcessToggleKey(Opt.ScanSymlink);    return TRUE;
          case VK_F7: ProcessToggleKey(Opt.ShowFolders, 3); return TRUE;
        }
        break;

    }
  }
  return FALSE;
}
#else
int BranchPanel::ProcessKey(const INPUT_RECORD *Rec)
{
  if ( Rec->EventType == KEY_EVENT || Rec->EventType == FARMACRO_KEY_EVENT )
  {
    if ( Rec->Event.KeyEvent.bKeyDown )
    {
      DWORD ControlState = Rec->Event.KeyEvent.dwControlKeyState & (LEFT_ALT_PRESSED|LEFT_CTRL_PRESSED|RIGHT_ALT_PRESSED|RIGHT_CTRL_PRESSED|SHIFT_PRESSED);
      if ( ControlState & RIGHT_ALT_PRESSED )
      {
        ControlState &= ~RIGHT_ALT_PRESSED;
        ControlState |= LEFT_ALT_PRESSED;
      }
      for ( WORD i = 0 ; i < Rec->Event.KeyEvent.wRepeatCount ; i++ )
      {
        switch ( ControlState )
        {
          case 0:
            switch ( Rec->Event.KeyEvent.wVirtualKeyCode )
            {
              case VK_F1: Info.ShowHelp(Info.ModuleName, HlfId.Panel, FHELP_USECONTENTS|FHELP_NOSHOWERROR); return TRUE;
              case VK_F7: ProcessRemoveKey();                                                               return TRUE;
            }
            break;
          case LEFT_ALT_PRESSED|SHIFT_PRESSED:
            switch ( Rec->Event.KeyEvent.wVirtualKeyCode )
            {
              case VK_F3: ProcessGoToKey();                     return TRUE;
              case VK_F5: ProcessToggleKey(Opt.FullPath, 3);    return TRUE;
              case VK_F6: ProcessToggleKey(Opt.ScanSymlink);    return TRUE;
              case VK_F7: ProcessToggleKey(Opt.ShowFolders, 3); return TRUE;
            }
            break;
        }
      }
    }
  }
  return FALSE;
}
#endif

void GoToFile(const TCHAR *Target, BOOL AnotherPanel)
{
  HANDLE _PANEL_HANDLE = AnotherPanel ? PANEL_PASSIVE : PANEL_ACTIVE;
  PanelRedrawInfo PRI;
  PanelInfo PInfo;
#ifndef FAR2
  PInfo.StructSize = sizeof(PanelInfo);
#endif
  const TCHAR *p = FSF.PointToName(const_cast<TCHAR*>(Target));
  TCHAR Name[512], Dir[512] = _T("");
  lstrcpy(Name, p);
  int pathlen = (int)(p-Target);
  if ( pathlen )
    lstrcpyn(Dir, Target, pathlen);
  Dir[pathlen] = _T('\0');
  if ( *Dir )
  {
#ifdef FAR2
    Info.PanelControl(_PANEL_HANDLE,FCTL_SETPANELDIR,0,(PANELCONTROLPARAM2)Dir);
#else
    FarPanelDirectory fpd = { sizeof(FarPanelDirectory), Dir, NULL, {0}, NULL };
    Info.PanelControl(_PANEL_HANDLE,FCTL_SETPANELDIRECTORY,0,(PANELCONTROLPARAM2)&fpd);
#endif
  }
  Info.PanelControl(_PANEL_HANDLE,FCTL_GETPANELINFO,0,(PANELCONTROLPARAM2)&PInfo);
  PRI.CurrentItem = PInfo.CurrentItem;
  PRI.TopPanelItem = PInfo.TopPanelItem;
  for ( int J = 0 ; J < (int)PInfo.ItemsNumber ; J++ )
  {
    bool found = false;
    size_t sz = Info.PanelControl(_PANEL_HANDLE,FCTL_GETPANELITEM,J,0);
    PluginPanelItem* PPI = (PluginPanelItem*)malloc(sz);
    if ( PPI )
    {
#ifdef FAR2
      Info.PanelControl(_PANEL_HANDLE,FCTL_GETPANELITEM,J,(PANELCONTROLPARAM2)PPI);
#else
      FarGetPluginPanelItem FPPI={sizeof(FarGetPluginPanelItem), sz, PPI};
      Info.PanelControl(_PANEL_HANDLE,FCTL_GETPANELITEM,J,(PANELCONTROLPARAM2)&FPPI);
#endif
      found = !FSF.LStricmp(Name, FSF.PointToName(piGetFileName(PPI)));
      free(PPI);
    }
    if ( found )
    {
      PRI.CurrentItem = J;
      PRI.TopPanelItem = J;
      break;
    }
  }
  Info.PanelControl(_PANEL_HANDLE,FCTL_REDRAWPANEL,0,(PANELCONTROLPARAM2)&PRI);
}

void BranchPanel::ProcessGoToKey()
{
  struct PanelInfo PInfo;
#ifndef FAR2
  PInfo.StructSize = sizeof(PanelInfo);
#endif
  Info.PanelControl(this,FCTL_GETPANELINFO,0,(PANELCONTROLPARAM2)&PInfo);
  TCHAR CurFileName[512] = _T("..");
  DWORD attributes = 0;
  size_t sz = Info.PanelControl(this,FCTL_GETPANELITEM,PInfo.CurrentItem,0);
  PluginPanelItem* PPI=(PluginPanelItem*)malloc(sz);
  if ( PPI )
  {
#ifdef FAR2
    Info.PanelControl(this,FCTL_GETPANELITEM,PInfo.CurrentItem,(PANELCONTROLPARAM2)PPI);
#else
    FarGetPluginPanelItem FPPI = {sizeof(FarGetPluginPanelItem), sz, PPI };
    Info.PanelControl(this,FCTL_GETPANELITEM,PInfo.CurrentItem,(PANELCONTROLPARAM2)&FPPI);
#endif
    attributes = piGetFileAttributes(PPI);
    lstrcpy(CurFileName, piGetFileName(PPI));
    free(PPI);
  }
  if ( lstrcmp(CurFileName, _T("..")) )
  {
    if ( attributes & FILE_ATTRIBUTE_DIRECTORY )
    {
#ifdef FAR2
      Info.PanelControl(PANEL_PASSIVE, FCTL_SETPANELDIR,0,(PANELCONTROLPARAM2)CurFileName);
#else
      FarPanelDirectory fpd = { sizeof(FarPanelDirectory), CurFileName, NULL, {0}, NULL };
      Info.PanelControl(PANEL_PASSIVE,FCTL_SETPANELDIRECTORY,0,(PANELCONTROLPARAM2)&fpd);
#endif
    }
    else
      GoToFile(CurFileName, true);
  }
  Info.PanelControl(PANEL_PASSIVE, FCTL_REDRAWPANEL, 0, NULL);
}

void BranchPanel::ProcessToggleKey(int& param, int limit)
{
  param = ++param % limit;
  Info.PanelControl(this,FCTL_UPDATEPANEL,0,NULL);
  Info.PanelControl(this,FCTL_REDRAWPANEL,0,NULL);
}

void BranchPanel::ProcessRemoveKey()
{
  bool pack = false;
  size_t shSize = sizeof(struct ex(FarMenuItem))*(rootItemsNumber+2);
  struct ex(FarMenuItem) *shMenu = (struct ex(FarMenuItem) *)malloc(shSize);
  memset(shMenu, 0, shSize);
  for ( int i = 0 ; i < rootItemsNumber ; i++ )
  {
    shMenu[i].Text = rootList[i].Name;
    if ( rootList[i].State == riActive )
      shMenu[i].Flags = MIF_CHECKED;
    else
      shMenu[i].Flags = 0;
  }
  shMenu[rootItemsNumber].Flags = MIF_SEPARATOR;
  shMenu[rootItemsNumber+1].Text = GetMsg(MConfigAdv);
#ifdef FAR2
  const int BreakKeys[] = { VK_INSERT, VK_SPACE, VK_DELETE, VK_RETURN, 0 };
#else
  const FarKey BreakKeys[] = { { VK_INSERT, 0 }, { VK_SPACE, 0 }, { VK_DELETE, 0 }, { VK_RETURN, 0 }, { 0, 0 } };
#endif
  intptr_t BreakCode = -1, Ret = 0;
  for ( ; ; )
  {
    for ( int i = 0 ; i < rootItemsNumber+2 ; i++ )
    {
      if ( i == Ret )
        shMenu[i].Flags |= MIF_SELECTED;
      else
        shMenu[i].Flags &= ~MIF_SELECTED;
    }
    Ret = INFOMENU(-1,-1,0,FMENU_WRAPMODE,GetMsg(MEdListTitle),GetMsg(MEdListBottom),HlfId.EditList,BreakKeys,&BreakCode,shMenu,rootItemsNumber+2);
    if ( Ret >= 0 )
    {
      if ( Ret == rootItemsNumber+1 )
      {
        if ( ( BreakCode < 0 || BreakCode == 3 ) && ConfigAdv() )
        {
          Info.PanelControl(this,FCTL_UPDATEPANEL,0,NULL);
          Info.PanelControl(this,FCTL_REDRAWPANEL,0,NULL);
        }
        continue;
      }
      switch ( BreakCode )
      {
        case 0:
        case 1:
          switch ( rootList[Ret].State )
          {
            case riActive:
            case riDelete:
              rootList[Ret].State = riIgnore;
              shMenu[Ret].Flags &= ~MIF_CHECKED;
              break;
            case riIgnore:
              rootList[Ret].State = riActive;
              shMenu[Ret].Flags |= MIF_CHECKED;
              break;
          }
          if ( ++Ret >= rootItemsNumber )
            Ret = 0;
          continue;
        case 2:
          switch ( rootList[Ret].State )
          {
            case riActive:
            case riIgnore:
              rootList[Ret].State = riDelete;
              shMenu[Ret].Flags &= ~MIF_CHECKED;
              shMenu[Ret].Flags |= MIF_GRAYED;
              break;
            case riDelete:
              rootList[Ret].State = riIgnore;
              shMenu[Ret].Flags &= ~MIF_CHECKED;
              shMenu[Ret].Flags &= ~MIF_GRAYED;
              break;
          }
          if ( ++Ret >= rootItemsNumber )
            Ret = 0;
          continue;
        default:
          pack = true;
          break;
      }
    }
    break;
  }
  if ( pack )
    rootItemsNumber -= packRoot();
  free(shMenu);
  int activeNumber = 0;
  for ( int i = 0 ; i < rootItemsNumber+2 ; i++ )
    if ( rootList[i].State == riActive )
      activeNumber++;
  if ( !activeNumber )
  {
    Info.PanelControl(this, FCTL_CLOSEPANEL, 0, NULL);
    return;
  }
  if ( pack )
  {
    struct PanelInfo PInfo;
#ifndef FAR2
    PInfo.StructSize = sizeof(PanelInfo);
#endif
    Info.PanelControl(this,FCTL_UPDATEPANEL,0,NULL);
    Info.PanelControl(this,FCTL_REDRAWPANEL,0,NULL);
    Info.PanelControl(PANEL_PASSIVE,FCTL_GETPANELINFO,0,(PANELCONTROLPARAM2)&PInfo);
    if ( PInfo.PanelType == PTYPE_QVIEWPANEL )
    {
      Info.PanelControl(PANEL_PASSIVE,FCTL_UPDATEPANEL,0,NULL);
      Info.PanelControl(PANEL_PASSIVE,FCTL_REDRAWPANEL,0,NULL);
    }
  }
}
