#include <windows.h>
#include "plugin.hpp"
#include "version.hpp"
#include <PluginSettings.hpp>
#include "CRT/crt.hpp"
#include "TrueBranch.hpp"
#include "TrueOpt.hpp"

//#include "_syslog.hpp"

struct HELPIDS HlfId =
{
  _T("Panel"),
  _T("Config"),
  _T("Init"),
  _T("EditList"),
  _T("WarnGetInfo"),
  _T("ConfigAdv")
};

struct RegistryStr REGStr =
{
  _T("Add2PlugMenu"),
  _T("Add2DisksMenu"),
  _T("ScanMode"),
  _T("ScanFolders"),
  _T("FilesMask"),
  _T("ESCConfirm"),
  _T("AutoSelected"),
  _T("FullPath"),
  _T("WarnGetInfo"),
  _T("ScanSymlink")
};

struct Options Opt =
{
  1,        // Add2PlugMenu
  0,        // Add2DisksMenu
  0,        // DisksMenuDigit
  0,        // ScanMode"),
  0,        // ScanFolders
 _T("*.*"), // FilesMask
  1,        // ESCConfirm
  0,        // AutoSelected
  0,        // FullPath
  1,        // WarnGetInfo
  0         // ScanSymlink
};

static TCHAR *GetCommaWord(TCHAR *Src, TCHAR *Word, const TCHAR *Separators)
{
  int WordPos, SkipBrackets = 0;
  if ( !*Src )
    return NULL;
  for ( WordPos = 0 ; *Src ; Src++,WordPos++ )
  {
    switch ( SkipBrackets )
    {
      case 0:
        if ( *Src == _T('[') && wcschr(Src+1, _T(']')) != NULL )
          SkipBrackets = 2;
        if ( *Src == _T('\"') && wcschr(Src+1, _T('\"')) != NULL )
          SkipBrackets = 1;
        break;
      case 1:
        if ( *Src == _T('\"') )
          SkipBrackets = 0;
        break;
      case 2:
        if ( *Src == _T(']') )
          SkipBrackets = 0;
        break;
    }
    if ( SkipBrackets == 0 && wcschr(Separators, *Src) )
    {
      Word[WordPos] = 0;
      Src++;
      while ( _istspace(*Src) )
        Src++;
      return Src;
    }
    else
      Word[WordPos] = *Src;
  }
  Word[WordPos] = 0;
  return Src;
}

bool allFiles(const TCHAR* Mask)
{
  return !lstrcmp(Mask, _T("*.*")) || !lstrcmp(Mask, _T("*"));
}

static bool compactMask(TCHAR *str, bool FullPath)
{
  if ( *str )
  {
    TCHAR *p;
    while ( ( p = wcsstr(str, _T("**")) ) != NULL )
      lstrcpy(p, p+1);
    if ( FullPath )
    {
      TCHAR w[512];
      TCHAR *bak = wcsdup(str);
      TCHAR *s = bak;
      *str = 0;
      while ( ( s = GetCommaWord(s, w, _T(",;")) ) != NULL )
      {
        FSF.Trim(w);
        bool quote = *w == _T('\"');
        FSF.Unquote(w);
        if ( *w )
        {
          if ( allFiles(w) )
          {
            lstrcpy(str, _T("*"));
            return true;
          }
          if ( *str )
            lstrcat(str, _T(","));
          if ( quote )
            lstrcat(str, _T("\""));
          if ( !wcsstr(w, _T("\\")) )
            lstrcat(str, _T("*\\"));
          lstrcat(str, w);
          if ( quote )
            lstrcat(str, _T("\""));
        }
      }
      free(bak);
    }
  }
  return false;
}

bool parseMasks(void)
{
  bool FullPath = wcschr(Opt.FilesMask, _T('\\')) != NULL;
  if ( wcschr(Opt.FilesMask, _T('/')) )
    Opt.FarSearch = true;
  else if ( !FullPath )
    Opt.FarSearch = true;
  else
  {
    Opt.FarSearch = false;
    TCHAR *p = wcschr(lstrcpy(Opt.IncludeMask, Opt.FilesMask), _T('|'));
    if ( p )
    {
      lstrcpy(Opt.ExcludeMask, p+1);
      *p = 0;
    }
    else
      *Opt.ExcludeMask = 0;
    compactMask(Opt.IncludeMask, FullPath);
    return compactMask(Opt.ExcludeMask, FullPath);
  }
  return false;
}

void GetOpt(void)
{
  PLUGINSETTINGS(settings);
  Opt.Add2PlugMenu   = settings.Get(PLUGINHKEY, REGStr.Add2PlugMenu,   1);
  Opt.Add2DisksMenu  = settings.Get(PLUGINHKEY, REGStr.Add2DisksMenu,  0);
  Opt.ScanMode       = settings.Get(PLUGINHKEY, REGStr.ScanMode,       0);
  Opt.ShowFolders    = settings.Get(PLUGINHKEY, REGStr.ShowFolders,    0);
  Opt.ESCConfirm     = settings.Get(PLUGINHKEY, REGStr.ESCConfirm,     1);
  Opt.AutoSelected   = settings.Get(PLUGINHKEY, REGStr.AutoSelected,   0);
  Opt.FullPath       = settings.Get(PLUGINHKEY, REGStr.FullPath,       0);
  Opt.WarnGetInfo    = settings.Get(PLUGINHKEY, REGStr.WarnGetInfo,    1);
  Opt.ScanSymlink    = settings.Get(PLUGINHKEY, REGStr.ScanSymlink,    0);
  settings.Get(PLUGINHKEY, REGStr.FilesMask, Opt.FilesMask, ARRAYSIZE(Opt.FilesMask), _T("*.*"));
  parseMasks();
}

void SetOpt(enOpt what)
{
  PLUGINSETTINGS(settings);
  if ( what & enoScanMode )
    settings.Set(PLUGINHKEY, REGStr.ScanMode,       Opt.ScanMode);
  if ( what & enoWarnGetInfo )
    settings.Set(PLUGINHKEY, REGStr.WarnGetInfo,    Opt.WarnGetInfo);
  if ( what & enoOther )
  {
    settings.Set(PLUGINHKEY, REGStr.Add2PlugMenu,   Opt.Add2PlugMenu);
    settings.Set(PLUGINHKEY, REGStr.Add2DisksMenu,  Opt.Add2DisksMenu);
    settings.Set(PLUGINHKEY, REGStr.ShowFolders,    Opt.ShowFolders);
    settings.Set(PLUGINHKEY, REGStr.FilesMask,      Opt.FilesMask);
    settings.Set(PLUGINHKEY, REGStr.ESCConfirm,     Opt.ESCConfirm);
    settings.Set(PLUGINHKEY, REGStr.AutoSelected,   Opt.AutoSelected);
    settings.Set(PLUGINHKEY, REGStr.ScanSymlink,    Opt.ScanSymlink);
    settings.Set(PLUGINHKEY, REGStr.FullPath,       Opt.FullPath);
  }
}
